# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

module ILO_SDK
  # Contains helper methods for Chassis actions
  module ChassisHelper
    # Get the power metrics
    # @raise [RuntimeError] if the request failed
    # @return [Hash] power_metrics
    def get_power_metrics
      chassis = rest_get('/redfish/v1/Chassis')
      chassis_uri = response_handler(chassis)['links']['Member'][0]['href']
      power_metrics_uri = response_handler(rest_get(chassis_uri))['links']['PowerMetrics']['href']
      response = rest_get(power_metrics_uri)
      metrics = response_handler(response)
      power_supplies = []
      metrics['PowerSupplies'].each do |ps|
        power_supply = {
          'LineInputVoltage' => ps['LineInputVoltage'],
          'LineInputVoltageType' => ps['LineInputVoltageType'],
          'PowerCapacityWatts' => ps['PowerCapacityWatts'],
          'PowerSupplyType' => ps['PowerSupplyType'],
          'Health' => ps['Status']['Health'],
          'State' => ps['Status']['State']
        }
        power_supplies.push(power_supply)
      end
      {
        @host => {
          'PowerCapacityWatts' => metrics['PowerCapacityWatts'],
          'PowerConsumedWatts' => metrics['PowerConsumedWatts'],
          'PowerSupplies' => power_supplies
        }
      }
    end

    # Get the thermal metrics
    # @raise [RuntimeError] if the request failed
    # @return [Hash] thermal_metrics
    def get_thermal_metrics
      chassis = rest_get('/redfish/v1/Chassis/')
      chassis_uri = response_handler(chassis)['links']['Member'][0]['href']
      thermal_metrics_uri = response_handler(rest_get(chassis_uri))['links']['ThermalMetrics']['href']
      response = rest_get(thermal_metrics_uri)
      temperatures = response_handler(response)['Temperatures']
      temp_details = []
      temperatures.each do |temp|
        temp_detail = {
          'PhysicalContext' => temp['PhysicalContext'],
          'Name' => temp['Name'],
          'CurrentReading' => temp['ReadingCelsius'],
          'CriticalThreshold' => temp['LowerThresholdCritical'],
          'Health' => temp['Status']['Health'],
          'State' => temp['Status']['State']
        }
        temp_details.push(temp_detail)
      end
      { @host => temp_details }
    end

    # Get the temperature metrics ilo 5
    # @raise [RuntimeError] if the request failed
    # @return [Hash] thermal_metrics
    def get_temperature_metrics_5
      chassis = rest_get('/redfish/v1/Chassis/1/Thermal/')
      temperatures = response_handler(chassis)['Temperatures']
      temp_details = []
      temperatures.each do |temp|
        temp_detail = {
          'PhysicalContext' => temp['PhysicalContext'],
          'Name' => temp['Name'],
          'CurrentReading' => temp['ReadingCelsius'],
          'CriticalThreshold' => temp['UpperThresholdCritical'],
          'Health' => temp['Status']['Health'],
          'State' => temp['Status']['State']
        }
        temp_details.push(temp_detail)
      end
      temp_details
    end

    # Get the fan metrics ilo 5
    # @raise [RuntimeError] if the request failed
    # @return [Hash] thermal_metrics
    def get_fan_metrics_5
      chassis = rest_get('/redfish/v1/Chassis/1/Thermal/')
      fans = response_handler(chassis)['Fans']
      fan_details = []
      fans.each do |fan|
        fan_detail = {
          'ReadingUnits' => fan['ReadingUnits'],
          'Name' => fan['Name'],
          'CurrentReading' => fan['Reading'],
          'Health' => fan['Status']['Health'],
          'State' => fan['Status']['State']
        }
        fan_details.push(fan_detail)
      end
      fan_details
    end

    # Get the power metrics ilo 5
    # @raise [RuntimeError] if the request failed
    # @return [Hash] thermal_metrics
    def get_power_metrics_5
      chassis = rest_get('/redfish/v1/Chassis/1/Power/')
      power = response_handler(chassis)["PowerSupplies"]
      power_details = []
      power.each do |ps|
        power_detail =  {
          'LastPowerOutputWatts' => ps['LastPowerOutputWatts'],
          'PowerCapacityWatts' => ps['PowerCapacityWatts'],
          'Model' => ps['Model'],
          'Name' => ps['Name'],
          'Health' => ps['Status']['Health']
        }
        power_details.push(power_detail)
      end
      power_details
    end

    # Get the memory metrics ilo 5
    # @raise [RuntimeError] if the request failed
    # @return [Hash] thermal_metrics
    def get_memory_metrics_5
      chassis = rest_get("/redfish/v1/Systems/1/Memory/")
      memory = response_handler(chassis)['Members']
      memory_details = []

      memory.each do |mem|
        chassis = rest_get(mem['@odata.id'])
        memory = response_handler(chassis)
        memory_detail =  {
          'Health' => memory['Status']['Health']
        }
        memory_details.push(memory_detail)
      end
      memory_details
    end

    # Get the System metrics ilo 5
    # @raise [RuntimeError] if the request failed
    # @return [Hash] thermal_metrics
    def get_system_metrics_5
      chassis = rest_get("/redfish/v1/")
      system1 = response_handler(chassis)['Oem']['Hpe']['System']
      system2 = response_handler(chassis)['Oem']['Hpe']['Manager']
      system_details = []
      system_detail = ""

      system1.each do |sys1|
        system2.each do |sys2|

        system_detail =  {
          'System_Health' => sys1['Status']['Health'],
          'iLO_Health' => sys2['Status']['Health']
        }

       end
      end
      system_details.push(system_detail)
      system_details
    end

    # Get the Spec_System metrics ilo 5
    # @raise [RuntimeError] if the request failed
    # @return [Hash] thermal_metrics
    def get_system_spec_metrics_5
      chassis = rest_get('/redfish/v1/Systems/1/')
      system = response_handler(chassis)['Oem']['Hpe']['AggregateHealthStatus']
      system_details = []

      system_detail =  {
        'Bios_Health' => system['BiosOrHardwareHealth']['Status']['Health'],
        'Fan_Health' => system['Fans']['Status']['Health'],
        'Storage_Health' => system['Storage']['Status']['Health'],
        'Temp_Health' => system['Temperatures']['Status']['Health']
      }

      system_details.push(system_detail)
      system_details
    end

  end
end
